/*
 * (C) Copyright IBM Corp. 2003    All Rights Reserved
 * mdmpd.c ($Revision: 1.5 $)
 * US Government Users Restricted Rights -
 * Use, duplication or disclosure restricted by
 * GSA ADP Schedule Contract with IBM Corp.
 *
 * Author(s): S. Bader <shbader@de.ibm.com>
 *
 * This file is part of MDMPD.
 *
 * MDMPD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * MDMPD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with MDMPD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <sys/poll.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/types.h>

#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>

#include <mdstat.h>
#include <mdctrl.h>

#ifndef MAX
#define MAX(x, y)	((x > y) ? x : y)
#endif
#define ERR(t, ...)	fprintf(stderr, t, ##__VA_ARGS__)
#if DEBUG > 4
#define DBG(t, ...)	fprintf(stderr, "%s @ %d: " t, __FUNCTION__ , \
				__LINE__ , ##__VA_ARGS__)
#else
#define DBG(t, ...)
#endif

/* License statement */
static char *license = "\n(C) Copyright IBM Corp. 2003    All Rights Reserved.\nmdmpd.c ($Revision: 1.5 $)\nUS Government Users Resticted Rights -\nUse, Duplication or disclosure restricted by\nGSA ADP Schedule Contract with IBM Corp\n";

/* Global md data to share between all functions. */
struct mds_info		mdi1;
struct mds_info		mdi2;

/* Process/thread syncronization. */
int			shutdown	= 0;
pthread_mutex_t		access_mdstat	= PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t		access_runprobe	= PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t		runprobe	= PTHREAD_COND_INITIALIZER;

int recover_array(struct mds_array *mda) {
	int 	i;
	int	rc;

	rc = 0;
	for(i = 0; i < MAX_MD_DISKS && !shutdown; i++) {
		if(mda->disk[i].status == disk_missing)
			continue;
		if(mda->disk[i].status == disk_failed) {
			DBG("Probing md%d disk%d\n", mda->nr, i);
			rc++;
			pthread_mutex_unlock(&access_mdstat);
			if(md_poll_disk(mda, i) == 0) {
				DBG("Disk ok\n");
				rc--;
			} else {
				DBG("DISK still failing\n");
			}
			pthread_mutex_lock(&access_mdstat);
		}
	}
	return rc;
}

void *recover_arrays(void *data) {
	int		i;
	int		failed;
	struct timespec	ts;
	sigset_t	sigmask;

	DBG("Thread starts\n");

	sigfillset(&sigmask);
	pthread_sigmask(SIG_BLOCK, &sigmask, NULL);

	do {
		pthread_mutex_lock(&access_mdstat);
		failed = 0;
		for(i = 0; i < 256 && !shutdown; i++) {
			if(mdi1.md[i].level != multipath)
				continue;
			if(mdi1.md[i].failed < 1)
				continue;

			failed += recover_array(&mdi1.md[i]);
		}
		pthread_mutex_unlock(&access_mdstat);

		pthread_mutex_lock(&access_runprobe);
		if(!failed) {
			DBG("Thread waits for failed devices.\n");
			pthread_cond_wait(&runprobe, &access_runprobe);
		} else {
			DBG("Thread sleeps for next cycle\n");
			ts.tv_sec  = time(NULL) + 15;
			ts.tv_nsec = 0;
			pthread_cond_timedwait(&runprobe,&access_runprobe,&ts);
		}
		pthread_mutex_unlock(&access_runprobe);
		DBG("Thread wakes up\n");
	} while(!shutdown);

	DBG("Thread stops\n");

	return NULL;
}

int find_disk_changes(struct mds_array *md1, struct mds_array *md2) {
	int i;
	int rc;

	rc = 0;
	for(i = 0; i < MAX_MD_DISKS; i++) {
		if(
			md2->disk[i].status != disk_missing &&
			md1->disk[i].status == disk_missing
		) {
			md_put_disk(md2, i);
			continue;
		}

		if(
			md1->disk[i].major != md2->disk[i].major ||
			md1->disk[i].minor != md2->disk[i].minor
		) {
			if(md2->disk[i].status != disk_missing)
				md_put_disk(md2, i);
			md_get_disk(md1, i);
		}

		if(md1->disk[i].status == disk_failed) {
			rc++;
		}
	}

	return rc;
}

void find_md_changes(void) {
	int			i;
	int			failcount;

	failcount = 0;
	for(i = 0; i < 256; i++) {
		if(mdi1.md[i].status != mdi2.md[i].status) {
			if(mdi1.md[i].status == array_active)
				md_get_array(&mdi1.md[i]);
			else
				md_put_array(&mdi2.md[i]);
		}
		failcount += find_disk_changes(&mdi1.md[i], &mdi2.md[i]);
	}

	if(failcount) {
		pthread_mutex_lock(&access_runprobe);
		pthread_cond_signal(&runprobe);
		pthread_mutex_unlock(&access_runprobe);
	}
	memcpy(&mdi2, &mdi1, sizeof(mdi2));
}

void poll_status(void) {
	do {
		if(mdstat_wait(5000) == 0) {
			pthread_mutex_lock(&access_mdstat);
			mdstat_get(&mdi1);
			DBG("new events %ld\n", mdi1.events);
			find_md_changes();
			pthread_mutex_unlock(&access_mdstat);
		}
	} while(!shutdown);
}

void sighandler(int signal) {
	DBG("Catched signal %d\n", signal);
	shutdown = 1;
}

int main(int argc, char **argv)
{
	pthread_t		recover_arrays_thread;
	struct sigaction	action;

	(void) license;

	if(mdstat_register(MDSTAT_TYPE_WITH_EVENTS) != 0) {
		if(mdstat_register(MDSTAT_TYPE_WITHOUT_EVENTS) != 0) {
			ERR("Unable to open /proc/mdstat\n");
			return -1;
		}
	}

	if(make_node_directory() < 0) {
		ERR("Couldn't create node directory.\n");
		return -1;
	}

	DBG("Data size = %ld\n", sizeof(mdi1));
	memset(&mdi2, 0, sizeof(mdi2));
	mdstat_get(&mdi1);

	if(mdi1.events < 0) {
		DBG("Kernel md module does not support events\n");
		mdstat_unregister();
		if(mdstat_register(MDSTAT_TYPE_WITHOUT_EVENTS) != 0) {
			ERR("Failed to register with non-event filehandle\n");
			return -1;
		}
	}
	DBG("There were %ld events\n", mdi1.events);

	find_md_changes();

	if(daemon(0, 0) < 0) {
		ERR("Failed to daemonise (errno=%d).\n", errno);
		return -1;
	}

	action.sa_handler = sighandler;
	sigaction(SIGHUP,  &action, NULL);
	sigaction(SIGINT,  &action, NULL);
	sigaction(SIGQUIT, &action, NULL);
	sigaction(SIGTERM, &action, NULL);

	pthread_create(&recover_arrays_thread, NULL, recover_arrays, NULL);

	mlockall(MCL_CURRENT);

	poll_status();

	/* Shutdown and wait for polling thread. */
	pthread_mutex_lock(&access_runprobe);
	pthread_cond_signal(&runprobe);
	pthread_mutex_unlock(&access_runprobe);
	pthread_join(recover_arrays_thread, NULL);

	memset(&mdi1, 0, sizeof(mdi1));
	find_md_changes();

	munlockall();

	DBG("Ending mdstat monitor\n");
	mdstat_unregister();

	return 0;
}

